# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The XHTML renderer

This renderer is dedicated to the Nagare framework
"""

from __future__ import with_statement

import operator
import types
# import urllib
import imghdr

from lxml import etree as ET
import peak.rules
import webob

from nagare import security, ajax, presentation, partial

from nagare.namespaces import xml
from nagare.namespaces.xml import TagProp
from nagare.namespaces import xhtml_base


# ---------------------------------------------------------------------------

# Generic methods to add an attribute to a tag
# --------------------------------------------

@peak.rules.when(xml.add_attribute, (xml._Tag, basestring, ajax.Update))
def add_ajax_attribute(self, name, value):
    """Add an attribute with a ``ajax.Update`` value

    In:
      - ``self`` -- the tag
      - ``name`` -- name of the attribute
      - ``value`` -- ``ajax.Update`` value
    """
    # Generate a XHR request
    xml.add_attribute(self, name, value.generate_action(self._actions[0], self.renderer))


@peak.rules.when(xml.add_attribute, (xml._Tag, basestring, types.FunctionType))
def add_function_attribute(self, name, value):
    """Add an attribute with a function value

    In:
      - ``self`` -- the tag
      - ``name`` -- name of the attribute
      - ``value`` -- function value
    """
    # Transcode the function to javascript
    xml.add_attribute(self, name, ajax.JS(value))


@peak.rules.when(xml.add_attribute, (xml._Tag, basestring, types.MethodType))
def add_method_attribute(self, name, value):
    """Add an attribute with a method value

    In:
      - ``self`` -- the tag
      - ``name`` -- name of the attribute
      - ``value`` -- method value
    """
    # Transcode the method to javascript
    xml.add_attribute(self, name, ajax.JS(value))


@peak.rules.when(xml.add_attribute, (xml._Tag, basestring, ajax.JS))
def add_js_attribute(self, name, value):
    """Add an attribute with a ``ajax.JS`` value

    In:
      - ``self`` -- the tag
      - ``name`` -- name of the attribute
      - ``value`` -- ``ajax.JS`` value
    """
    self.set(name, value.generate_action(None, self.renderer))


# ---------------------------------------------------------------------------

def absolute_url(url, static):
    """Convert a relative URL of a static content to an absolute one

    In:
      - ``url`` -- url to convert
      - ``static`` -- URL prefix of the static contents

    Return:
      - an absolute URL
    """
    if url.startswith('#'):
        return url

    i = url.find(':')
    if ((i == -1) or not url[:i].isalpha()) and url and (url[0] != '/'):
        # If this is a relative URL, it's relative to the statics directory
        if not static.endswith('/') and not url.startswith('/'):
            url = static + '/' + url
        else:
            url = static + url

    return url


class HeadRenderer(xhtml_base.HeadRenderer):
    """The XHTML head Renderer

    This renderer knows about the static contents of the application
    """
    def __init__(self, static_url):
        """Renderer initialisation

        The ``HeadRenderer`` keeps track of the javascript and css used by every views,
        to be able to concatenate them into the ``<head>`` section.
        """
        super(HeadRenderer, self).__init__()

        # Directory where are located the static contents of the application
        self.static_url = static_url

        self._named_css = {}         # CSS code
        self._css_url = {}           # CSS URLs
        self._named_javascript = {}  # Javascript code
        self._javascript_url = {}    # Javascript URLs

        self._order = 0              # Memorize the order of the javascript and css

    def css(self, name, style, **kw):
        """Memorize an in-line named css style

        In:
          - ``name`` -- unique name of this css style (to prevent double definition)
          - ``style`` -- the css style
          - ``kw`` -- attributes of the generated ``<style>`` tag

        Return:
          - ``()``
        """
        self._named_css.setdefault(name, (self._order, style, kw))
        self._order += 1
        return ()

    def css_url(self, url, **kw):
        """Memorize a css style URL

        In:
          - ``url`` -- the css style URL
          - ``kw`` -- attributes of the generated ``<link>`` tag

        Return:
          - ``()``
        """
        self._css_url.setdefault(absolute_url(url, self.static_url), (self._order, kw))
        self._order += 1
        return ()

    def javascript(self, name, script, **kw):
        """Memorize an in-line named javascript code

        In:
          - ``name`` -- unique name of this javascript code (to prevent double definition)
          - ``script`` -- the javascript code
          - ``kw`` -- attributes of the generated ``<script>`` tag

        Return:
          - ``()``
        """
        if callable(script):
            # Transcode the function or the method to javascript code
            script = ajax.javascript(script)

        if isinstance(script, ajax.JS):
            # Transcoded javascript needs a helper
            self.javascript_url('/static/nagare/pyjslib.js')
            script = script.javascript

        self._named_javascript.setdefault(name, (self._order, script, kw))
        self._order += 1
        return ()

    def javascript_url(self, url, **kw):
        """Memorize a javascript URL

        In:
          - ``url`` -- the javascript URL
          - ``kw`` -- attributes of the the generated ``<script>`` tag

        Return:
          - ``()``
        """
        self._javascript_url.setdefault(absolute_url(url, self.static_url), (self._order, kw))
        self._order += 1
        return ()

    def _style(self, append, tag, style):
        append(tag, style)

    def _script(self, append, tag, script):
        append(tag, script)

    def _get_named_css(self):
        """Return the list of the in-line named css styles, sorted by order of insertion

        Return:
          - list of (name, css style, attributes)
        """

        return [(name, style, attributes) for (name, (order, style, attributes)) in sorted(self._named_css.items(), key=operator.itemgetter(1))]

    def _get_css_url(self):
        """Return the list of css URLs, sorted by order of insertion

        Return:
          - list of css (URLs, attributes)
        """
        return [(url, attributes) for (url, (order, attributes)) in sorted(self._css_url.items(), key=operator.itemgetter(1))]

    def _get_named_javascript(self):
        """Return the list of named javascript codes, sorted by order of insertion

        Return:
          - list of (name, javascript code, attributes)
        """
        return [(name, js, attributes) for (name, (order, js, attributes)) in sorted(self._named_javascript.items(), key=operator.itemgetter(1))]

    def _get_javascript_url(self):
        """Return the list of javascript URLs, sorted by order of insertion

        Return:
          - list of javascript (URLs, attributes)
        """
        return [(url, attributes) for (url, (order, attributes)) in sorted(self._javascript_url.items(), key=operator.itemgetter(1))]


@presentation.render_for(HeadRenderer)
def render(self, h, *args):
    """
    Generate the ``<head>`` tree

    In:
      - ``h`` -- *not used*

    Return:
      - the ``<head>`` tree
    """

    # Create the tags to include the CSS styles and the javascript codes

    head = self.root

    if isinstance(head, ET.ElementBase) and (head.tag == 'head'):
        # If a ``<head>`` tag already exist, take its content
        head = self.head(head[:], dict(head.attrib))
    else:
        head = self.head(head)

    head.extend([self.link(rel='stylesheet', type='text/css', href=url, **attributes) for (url, attributes) in self._get_css_url()])
    head.extend([self.script(type='text/javascript', src=url, **attributes) for (url, attributes) in self._get_javascript_url()])

    head.extend([self.style(css, type='text/css', **attributes) for (name, css, attributes) in self._get_named_css()])
    head.extend([self.script(js, type='text/javascript', **attributes) for (name, js, attributes) in self._get_named_javascript()])

    return head


# ----------------------------------------------------------------------------------

class _HTMLActionTag(xhtml_base._HTMLTag):
    """Base class of all the tags with a ``.action()`` method
    """
    @partial.max_number_of_args(2)
    def action(self, action, args, with_request=False, permissions=None, subject=None, **kw):
        """Register an action

        In:
          - ``action`` -- action
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        if isinstance(action, ajax.Update):
            self._async_action(self.renderer, action, with_request)
        else:
            # Wrap the ``action`` into a wrapper that will check the user permissions
            action = security.wrapper(action, permissions, subject or self.renderer.component())
            action = partial.Partial(action, *args, **kw)

            # Double dispatch with the renderer
            self.renderer.action(self, action, with_request)

        return self

    def sync_action(self, renderer, action, with_request):
        """Register a synchronous action

        In:
          - ``renderer`` -- the current renderer
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        self.set(self._actions[1], renderer.register_callback(self._actions[0], action, with_request))

    def async_action(self, renderer, action, with_request):
        """Register an asynchronous action

        In:
          - ``renderer`` -- the current renderer
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        if not isinstance(action, ajax.Update):
            action = ajax.Update(action=action, with_request=with_request)

        self.set(self._actions[2], action.generate_action(self._actions[0], renderer))
    _async_action = async_action


# ----------------------------------------------------------------------------------

class Form(xhtml_base._HTMLTag):
    """ The ``<form>`` tag
    """
    def init(self, renderer):
        """Initialisation

        In:
          - ``renderer`` -- the current renderer

        Return:
          - ``self``
        """
        super(Form, self).init(renderer)

        # Set the default attributes : send the form with a POST, in utf-8
        self.set('enctype', 'multipart/form-data')
        self.set('method', 'post')
        self.set('accept-charset', 'utf-8')

        self.set('action', '?')
        # self.append(renderer.input(name='_charset_', value='utf-8', type='hidden'))

        # Add into the form the hidden fields of the session and continuation ids
        self.renderer.add_sessionid_in_form(self)
        return self

    def add_child(self, child):
        """Add a child to this ``<form>`` tag

        Delete the existing ``<form>`` tags in the child tree
        """
        if isinstance(child, ET.ElementBase):
            for f in child.findall('.//form'):
                f.replace(f.getchildren()[1:])

        if isinstance(child, Form):
            child = child.getchildren()[1:]

        super(Form, self).add_child(child)

    def _action(self, action, priority, args, with_request, permissions, subject, **kw):
        """Register an action

        In:
          - ``action`` -- action
          - ``priority`` - ``action`` priority
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        action = security.wrapper(action, permissions, subject or self.renderer.component())
        action = partial.Partial(action, *args, **kw)

        # Generate a hidden field with the action attached
        self.append(self.renderer.div(self.renderer.input(type='hidden', name=self.renderer.register_callback(priority, action, with_request))))
        return self

    @partial.max_number_of_args(2)
    def pre_action(self, action, args, with_request=False, permissions=None, subject=None, **kw):
        """Register an action that will be executed **before** the actions of the
        form elements

        In:
          - ``action`` -- action
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        return self._action(action, 0, args, with_request, permissions, subject, **kw)

    @partial.max_number_of_args(2)
    def post_action(self, action, args, with_request=False, permissions=None, subject=None, **kw):
        """Register an action that will be executed **after** the actions of the
        form elements

        In:
          - ``action`` -- action
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response object be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        return self._action(action, 3, args, with_request, permissions, subject, **kw)


# ----------------------------------------------------------------------------------

class TextInput(_HTMLActionTag):
    """Dispatcher class for all the ``<input>`` tags
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action

    def __call__(self, *children, **attrib):
        t = attrib.get('type', 'text')
        if t == 'text':
            # ``<input>`` with ``type=text``
            return super(TextInput, self).__call__(*children, **attrib)

        element = self.renderer.makeelement(t + '_input')
        element.tag = self.tag
        if hasattr(self, '_authorized_attribs'):
            element._authorized_attribs = self._authorized_attribs

        return element(**dict(self.attrib))(*children, **attrib)


class TextArea(_HTMLActionTag):
    """ ``<textarea>`` tags
    """
    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action

    @classmethod
    def clean_input(cls, action, args, v, **kw):
        return action(*(args + (v.replace('\r', ''),)), **kw)

    @classmethod
    def clean_input_with_request(cls, action, args, request, response, v, **kw):
        return cls.clean_input(action, (request, response) + args, v, **kw)

    @partial.max_number_of_args(2)
    def action(self, action, args, with_request=False, permissions=None, subject=None, **kw):
        """Register an action


        In:
          - ``action`` -- action
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response object be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        # The content sent to the action will have the '\r' characters deleted
        if not isinstance(action, ajax.Update):
            f = self.clean_input_with_request if with_request else self.clean_input
            action = partial.Partial(f, action, args)
            args = ()

        return super(TextArea, self).action(action, with_request=with_request, permissions=permissions, subject=subject, *args, **kw)


class PasswordInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=password`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action


class RadioInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=radio`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (2, 'value', 'onclick')
    async_action = _HTMLActionTag.sync_action

    def selected(self, flag):
        """(de)Select the tag

        In:
          - ``flag`` -- boolean to deselect / select the tag

        Return:
          - ``self``
        """
        if 'checked' in self.attrib:
            del self.attrib['checked']

        if flag:
            self.set('checked', 'checked')
        return self


class CheckboxInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=checkbox`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onclick')
    async_action = _HTMLActionTag.sync_action

    def selected(self, flag):
        """(de)Select the tag

        In:
          - ``flag`` -- boolean to deselect / select the tag

        Return:
          - ``self``
        """
        if 'checked' in self.attrib:
            del self.attrib['checked']

        if flag:
            self.set('checked', 'checked')
        return self


class SubmitInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=submit`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (4, 'name', 'onclick')


class HiddenInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=hidden`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action


class FileInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=file`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action

    def init(self, renderer):
        """Initialisation

        In:
          - ``renderer`` -- the current renderer

        Return:
          - ``self``
        """
        super(FileInput, self).init(renderer)

        self.set('name', 'file')
        return self


class ImageInput(_HTMLActionTag):
    """ ``<input>`` tags with ``type=image`` attributes
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (5, 'name', 'onclick')

    def add_child(self, child):
        """Add attributes to the image input

        In:
          - ``child`` -- attributes dictionary
        """
        super(ImageInput, self).add_child(child)

        # If this is a relative URL, it's relative to the statics directory
        # of the application
        src = self.get('src')
        if src is not None:
            self.set('src', absolute_url(src, self.renderer.head.static_url))


# ----------------------------------------------------------------------------------

class Option(xhtml_base._HTMLTag):
    """ ``<options>`` tags
    """
    def selected(self, values):
        """(de)Select the tags

        In:
          - ``values`` -- name or list of names of the tags to select

        Return:
          - ``self``
        """
        if not isinstance(values, (list, tuple)):
            values = (values,)

        if 'selected' in self.attrib:
            del self.attrib['selected']

        if self.get('value') in map(unicode, values):
            self.set('selected', 'selected')

        return self


class Select(_HTMLActionTag):
    """ ``<select>`` tags
    """

    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (1, 'name', 'onchange')
    async_action = _HTMLActionTag.sync_action

    @classmethod
    def normalize_input(cls, action, args, v, **kw):
        return action(*(args + (v if isinstance(v, (list, tuple)) else (v,),)), **kw)

    @classmethod
    def normalize_input_with_request(cls, action, args, request, response, v, **kw):
        return cls.normalize_input(action, (request, response) + args, v, **kw)

    @partial.max_number_of_args(2)
    def action(self, action, args, with_request=False, permissions=None, subject=None, **kw):
        """Register an action

        In:
          - ``action`` -- action
          - ``args``, ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response object be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on

        Return:
          - ``self``
        """
        if (self.get('multiple') is not None) and not isinstance(action, ajax.Update):
            # If this is a multiple select, the value sent to the action will
            # always be a list, even if only 1 item was selected
            f = self.normalize_input_with_request if with_request else self.normalize_input
            action = partial.Partial(f, action, args)
            args = ()

        return super(Select, self).action(action, with_request=with_request, permissions=permissions, subject=subject, *args, **kw)


# ----------------------------------------------------------------------------------

class A(_HTMLActionTag):
    """ ``<a>`` tags
    """
    # Tuple:
    #   - action type
    #   - name of the attribute for the synchronous action
    #   - name of the attribute for the asynchronous action
    _actions = (41, None, 'onclick')

    def sync_action(self, renderer, action, with_request):
        """Register a synchronous action

        In:
          - ``renderer`` -- the current renderer
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        href = self.get('href', '').partition('#')
        self.set('href', renderer.add_sessionid_in_url(href[0], (renderer.register_callback(4, action, with_request),)) + href[1] + href[2])

    def async_action(self, renderer, action, with_request):
        """Register an asynchronous action

        In:
          - ``renderer`` -- the current renderer
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        super(A, self).async_action(renderer, action, with_request)
        self.set('href', '#')
    _async_action = async_action

    """
    def url(self, *args):
        url = '/'.join([urllib.quote_plus(unicode(u).encode('utf-8'), '') for u in args])

        href = self.get('href', '')

        if not href.startswith(('#', '?')):
            href = ''

        self.set('href', self._renderer.url + '/' + url + href)

        return self
    """


@peak.rules.when(xml.add_attribute, (A, basestring, basestring))
def add_attribute(next_method, self, name, value):
    if name == 'href':
        value = absolute_url(value, self.renderer.url)

    next_method(self, name, value)


# ----------------------------------------------------------------------------------

class Img(_HTMLActionTag):
    """ ``<img>`` tags
    """
    @classmethod
    def _set_content_type(cls, request, response, _action, with_request):
        """Generate the image and guess its format

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``action`` -- function to call to generate the image data
          - ``with_request`` -- will the request and response objects be passed to the action?

        Return:
          - new response object raised
        """
        e = webob.exc.HTTPOk(headerlist=[('Content-Type', '')])
        img = _action(request, e) if with_request else _action()
        e.body = img

        content_type = e.content_type
        if not content_type:
            # If no ``Content-Type`` is already set, use the ``imghdr`` module
            # to guess the format of the image
            content_type = 'image/' + (imghdr.what(None, img[:32]) or '*')
        e.content_type = content_type

        raise e

    def sync_action(self, renderer, action, with_request):
        """Register a synchronous action

        The action will have to return the image data

        In:
          - ``renderer`` -- the current renderer
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on
        """
        f = partial.Partial(self._set_content_type, _action=action, with_request=with_request)
        self.set('src', renderer.add_sessionid_in_url(sep=';') + ';' + renderer.register_callback(2, f, with_request=True))
    async_action = sync_action

    def add_child(self, child):
        """Add attributes to the image

        In:
          - ``child`` -- attributes dictionary
        """
        super(Img, self).add_child(child)

        # If this is a relative URL, it's relative to the statics directory
        # of the application
        src = self.get('src')
        if src is not None:
            self.set('src', absolute_url(src, self.renderer.head.static_url))


# ----------------------------------------------------------------------------------

class Label(xhtml_base._HTMLTag):
    """ ``<label>`` tags
    """
    def init(self, renderer):
        """Initialisation

        In:
          - ``renderer`` -- the current renderer

        Return:
          - ``self``
        """
        super(Label, self).init(renderer)

        # Generate a unique value for the ``for`` attribute
        self.set('for', renderer.generate_id('id'))
        return self


# ----------------------------------------------------------------------------------

class Script(xhtml_base._HTMLTag):
    pass


@peak.rules.when(xml.add_child, (xhtml_base._HTMLTag, Script))
def add_script_child(next_method, self, script):
    """Add a <script> to a tag

    In:
      - ``self`` -- the tag
      - ``script`` -- the script to add
    """
    self.renderer.head._script(next_method, self, script)
    return ''


class Style(xhtml_base._HTMLTag):
    pass


@peak.rules.when(xml.add_child, (xhtml_base._HTMLTag, Style))
def add_style_child(next_method, self, style):
    """Add a <style> to a tag

    In:
      - ``self`` -- the tag
      - ``style`` -- the style to add
    """
    self.renderer.head._style(next_method, self, style)
    return ''


# ----------------------------------------------------------------------------------

class Renderer(xhtml_base.Renderer):
    """The XHTML synchronous renderer
    """
    XML_DOCTYPE = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
    HTML_DOCTYPE = '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">'

    head_renderer_factory = HeadRenderer

    # Redefinition of the he HTML tags with actions
    # ---------------------------------------------

    a = TagProp('a', set(xhtml_base.allattrs + xhtml_base.focusattrs + ('charset', 'type', 'name', 'href', 'hreflang', 'rel', 'rev', 'shape', 'coords', 'target', 'oncontextmenu')), A)
    area = TagProp('area', set(xhtml_base.allattrs + xhtml_base.focusattrs + ('shape', 'coords', 'href', 'nohref', 'alt', 'target')), A)
    button = TagProp('button', set(xhtml_base.allattrs + xhtml_base.focusattrs + ('name', 'value', 'type', 'disabled')), SubmitInput)
    form = TagProp('form', set(xhtml_base.allattrs + ('action', 'method', 'name', 'enctype', 'onsubmit', 'onreset', 'accept_charset', 'target')), Form)
    img = TagProp('img', set(xhtml_base.allattrs + (
        'src', 'alt', 'name', 'longdesc', 'width', 'height', 'usemap', 'ismap'
        'align', 'border', 'hspace', 'vspace', 'lowsrc')
    ), Img)
    input = TagProp('input', set(xhtml_base.allattrs + xhtml_base.focusattrs + (
        'type', 'name', 'value', 'checked', 'disabled', 'readonly', 'size', 'maxlength', 'src'
        'alt', 'usemap', 'onselect', 'onchange', 'accept', 'align', 'border')
    ), TextInput)
    label = TagProp('label', set(xhtml_base.allattrs + ('for', 'accesskey', 'onfocus', 'onblur')), Label)
    option = TagProp('option', set(xhtml_base.allattrs + ('selected', 'disabled', 'label', 'value')), Option)
    select = TagProp('select', set(xhtml_base.allattrs + ('name', 'size', 'multiple', 'disabled', 'tabindex', 'onfocus', 'onblur', 'onchange', 'rows')), Select)
    textarea = TagProp('textarea', set(xhtml_base.allattrs + xhtml_base.focusattrs + ('name', 'rows', 'cols', 'disabled', 'readonly', 'onselect', 'onchange', 'wrap')), TextArea)

    script = TagProp('script', set(('id', 'charset', 'type', 'language', 'src', 'defer')), Script)
    style = TagProp('style', set(xhtml_base.i18nattrs + ('id', 'type', 'media', 'title')), Style)

    _specialTags = dict(
        text_input=TextInput,
        radio_input=RadioInput,
        checkbox_input=CheckboxInput,
        submit_input=SubmitInput,
        hidden_input=HiddenInput,
        file_input=FileInput,
        password_input=PasswordInput,
        image_input=ImageInput
    )

    @classmethod
    def class_init(cls, specialTags):
        """Class initialisation

        In:
          -- ``special_tags`` -- tags that have a special factory
        """
        class CustomLookup(ET.CustomElementClassLookup):
            def __init__(self, specialTags, defaultLookup):
                super(CustomLookup, self).__init__(defaultLookup)
                self._specialTags = specialTags

            def lookup(self, node_type, document, namespace, name):
                return self._specialTags.get(name)

        cls._specialTags.update(specialTags)

        cls._custom_lookup = CustomLookup(cls._specialTags, ET.ElementDefaultClassLookup(element=xhtml_base._HTMLTag))
        cls._html_parser = ET.HTMLParser()
        cls._html_parser.setElementClassLookup(cls._custom_lookup)

    def __init__(self, parent=None, session=None, request=None, response=None, static_url='', static_path='', url='/'):
        """Renderer initialisation

        In:
          - ``parent`` -- parent renderer
          - ``session`` -- the session object
          - ``request`` -- the request object
          - ``response`` -- the response object
          - ``static_url`` -- url of the static contents of the application
          - ``static_path`` -- path of the static contents of the application
          - ``url`` -- url prefix of the application
        """
        super(Renderer, self).__init__(parent, static_url=static_url)

        if parent is None:
            self.session = session
            self.request = request
            self.response = response
            self.static_path = static_path
            self.url = url
            self.component = None
            self.model = None
        else:
            self.session = parent.session
            self.request = parent.request
            self.response = parent.response
            self.static_path = parent.static_path
            self.url = parent.url
            self.component = parent.component
            self.model = parent.model

    def SyncRenderer(self, *args, **kw):
        """Create an associated synchronous HTML renderer

        Return:
          - a new synchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return self.__class__(*args, **kw)

    def AsyncRenderer(self, *args, **kw):
        """Create an associated asynchronous HTML renderer

        Return:
          - a new asynchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return AsyncRenderer(*args, **kw)

    def makeelement(self, tag):
        """Make a tag

        In:
          - ``tag`` -- name of the tag to create

        Return:
          - the new tag
        """
        return self._makeelement(tag, self._html_parser)

    def parse_xml(self, source, fragment=False, no_leading_text=False, **kw):
        """Parse a XML file

        In:
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a XML fragment i.e a XML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the XML to parsed begins by a text, this text is keeped
          - ``kw`` -- keywords parameters are passed to the XML parser

        Return:
          - the root element of the parsed XML, if ``fragment`` is ``False``
          - a list of XML elements, if ``fragment`` is ``True``
        """
        if isinstance(source, basestring):
            source = absolute_url(source, self.static_path)

        return super(Renderer, self).parse_xml(source, fragment, no_leading_text, **kw)

    def parse_html(self, source, fragment=False, no_leading_text=False, xhtml=False, **kw):
        """Parse a (X)HTML file

        In:
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a HTML fragment i.e a HTML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the HTML to parsed begins by a text, this text is keeped
          - ``xhtml`` -- is the HTML to parse a valid XHTML ?
          - ``kw`` -- keywords parameters are passed to the HTML parser

        Return:
          - the root element of the parsed HTML, if ``fragment`` is ``False``
          - a list of HTML elements, if ``fragment`` is ``True``
        """
        if isinstance(source, basestring):
            source = absolute_url(source, self.static_path)

        parser = ET.XMLParser(**kw) if xhtml else ET.HTMLParser(**kw)
        parser.setElementClassLookup(self._custom_lookup)

        return self._parse_html(parser, source, fragment, no_leading_text, **kw)

    @property
    def doctype(self):
        """Generate the DOCTYPE of the document

        If a doctype was set on the response object, use it
        Else, use the HTML ou XHTML doctypes of this renderer

        Return:
          - the doctype
        """
        response = self.response

        if response.doctype is not None:
            return response.doctype

        return self.XML_DOCTYPE if response.xml_output else self.HTML_DOCTYPE

    @property
    def content_type(self):
        """Generate the content type of the document

        If a content type was set on the response object, use it
        Else, use the HTML ou XHTML content type of this renderer

        Return:
          - the content type
        """
        response = self.response

        if response.content_type:
            return response.content_type

        return 'application/xhtml+xml' if response.xml_output else 'text/html'

    def start_rendering(self, component, model):
        """Method called before to render a component

        In:
          - ``component`` -- component to render
          - ``model`` -- name of the view to use
        """
        self.component = component
        self.model = model

        if component.url is not None:
            self.url += '/' + component.url

    def action(self, tag, action, with_request):
        """Register a synchronous action on a tag

        In:
          - ``tag`` -- the tag
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        tag.sync_action(self, action, with_request)

    def register_callback(self, priority, f, with_request, render=None):
        """Register an action

        Register the action to the current rendered component

        In:
          - ``priority`` - -priority of the action
          - ``f`` -- the action
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``render`` -- render method to generate the view after the ``f`` action will be called
        """
        if self.component is not None:
            return self.component.register_callback(self.model or None, priority, f, with_request, render)

        return ''

    def decorate_error(self, element, error):
        """During the rendering, highlight an element that has an error

        In:
          - ``element`` -- the element in error
          - ``error`` -- the error text
        """
        if error is None:
            return element

        div = self.div(class_='nagare-error-input')
        div.append(element)

        return self.div(
            div,
            self.div(error, class_='nagare-error-message'),
            class_='nagare-error-field'
        )

    def add_sessionid_in_form(self, form):
        """Add the session and continuation ids into a ``<form>``

        Forward this call to the sessions manager

        In:
          - ``form`` -- the form tag
        """
        if self.session:
            form(self.div(self.session.sessionid_in_form(self, self.request, self.response), class_='nagare-generated nagare-session-data'))

    def add_sessionid_in_url(self, u='', params=None, sep='&'):
        """Add the session and continuation ids into an url

        Forward this call to the sessions manager

        In:
          - ``u`` -- the url
          - ``params`` -- query string of the url

        Return:
          - the completed url
        """
        i = u.find(':')
        if ((i == -1) or not u[:i].isalpha()) and (not u or (u[0] != '/')):
            u = self.url + '/' + u

        if params is None:
            params = ()

        if self.session:
            u += '?' + sep.join(self.session.sessionid_in_url(self.request, self.response) + params)

        return u


class AsyncHeadRenderer(HeadRenderer):
    def __init__(self, static_url):
        """Renderer initialisation

        The ``HeadRenderer`` keeps track of the javascript and css used by every views,
        to be able to concatenate them into the ``<head>`` section.
        """
        super(AsyncHeadRenderer, self).__init__(static_url=static_url)

        self._anonymous_css = []         # CSS
        self._anonymous_javascript = []  # Javascript code

    def _css(self, style):
        """Memorize an in-line anonymous css style

        In:
          - ``style`` -- the css style
        """
        self._anonymous_css.append((self._order, style))
        self._order += 1

    def _javascript(self, script):
        """Memorize an in-line anonymous javascript code

        In:
          - ``script`` -- the javascript code
        """
        self._anonymous_javascript.append((self._order, script))
        self._order += 1

    def _style(self, append, tag, style):
        self._css(style.text or '')

    def _script(self, append, tag, script):
        url = script.get('src')

        if url:
            self.javascript_url(url)
        else:
            self._javascript(script.text or '')

    def _get_anonymous_css(self):
        """Return the list of the in-line anonymous css styles, sorted by order of insertion

        Return:
          - list of css styles
        """
        return [css for (order, css) in sorted(self._anonymous_css)]

    def _get_anonymous_javascript(self):
        """Return the list of anonymous javascript codes, sorted by order of insertion

        Return:
          - list of javascript codes
        """
        return [js for (order, js) in sorted(self._anonymous_javascript)]


@presentation.render_for(AsyncHeadRenderer)
def render_async_head(self, h, *args):
    """Generate a javascript view of the head

    In:
      - ``h`` -- the current renderer

    Return:
      - a javascript string
    """
    return "nagare_loadAll(%s, %s, %s, %s, %s, %s)" % (
        ajax.py2js(self._get_named_css(), h),
        ajax.py2js(r'\n'.join(self._get_anonymous_css()), h),
        ajax.py2js(self._get_css_url(), h),
        ajax.py2js(self._get_named_javascript(), h),
        ajax.py2js(';'.join(self._get_anonymous_javascript()), h),
        ajax.py2js(self._get_javascript_url(), h)
    )


class AsyncRenderer(Renderer):
    """The XHTML asynchronous renderer
    """
    head_renderer_factory = AsyncHeadRenderer

    def __init__(self, parent=None, session=None, request=None, response=None, static_url='', static_path='', url='/', async_header=False):
        """Renderer initialisation

        In:
          - ``parent`` -- parent renderer
          - ``session`` -- the session object
          - ``request`` -- the request object
          - ``response`` -- the response object
          - ``static_url`` -- url of the static contents of the application
          - ``static_path`` -- path of the static contents of the application
          - ``url`` -- url prefix of the application
          - ``async_header`` -- is the head renderer to create a synchronous or an asynchronous one?
        """
        super(AsyncRenderer, self).__init__(parent, session, request, response, static_url, static_path, url)

        if not (parent or async_header):
            self.head = HeadRenderer(static_url=static_url)

        self.async_root = True
        self.wrapper_to_generate = False  # Add a ``<div>`` around the rendering ?

    def SyncRenderer(self, *args, **kw):
        """Create an associated synchronous HTML renderer

        Return:
          - a new synchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return Renderer(*args, **kw)

    def AsyncRenderer(self, *args, **kw):
        """Create an associated asynchronous HTML renderer

        Return:
          - a new asynchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return self.__class__(*args, **kw)

    def javascript_url(self, url):
        self.head.javascript_url(url)

    def _javascript(self, js):
        self.head._javascript(js)

    def _css(self, style):
        self.head._css(style)

    def action(self, tag, action, with_request):
        """Register an asynchronous action on a tag

        In:
          - ``tag`` -- the tag
          - ``action`` -- action
          - ``with_request`` -- will the request and response objects be passed to the action?
        """
        tag.async_action(self, action, with_request)

    def start_rendering(self, component, model):
        super(AsyncRenderer, self).start_rendering(component, model)
        self.async_root = False

    def end_rendering(self, output):
        """Method called after a component is rendered

        In:
          - ``output`` -- rendered tree

        Out:
          - rendered tree
        """
        if self.wrapper_to_generate:
            output = self.div(output, id=self.id, class_='nagare-generated nagare-async-view')

        return output

    def get_async_root(self):
        if not isinstance(self.parent, AsyncRenderer):
            # return None
            return self

        if self.parent.async_root:
            return self

        return self.parent.get_async_root()


# ---------------------------------------------------------------------------

if __name__ == '__main__':
    t = ((1, 'a'), (2, 'b'), (3, 'c'))

    h = Renderer()

    h.head << h.head.title('A test')
    h.head << h.head.javascript('__foo__', 'function() {}')
    h.head << h.head.meta(name='meta1', content='content1')

    with h.body(onload='javascript:alert()'):
        with h.ul:
            with h.li('Hello'):
                pass
            with h.li:
                h << 'world'
            h << h.li('yeah')

        with h.div(class_='foo'):
            with h.h1('moi'):
                h << h.i('foo')

        with h.div:
            h << 'yeah'
            for i in range(3):
                h << i

        with h.table(foo='foo'):
            for row in t:
                with h.tr:
                    for column in row:
                        with h.td:
                            h << column

    print h.html(presentation.render(h.head, None, None, None), h.root).write_htmlstring(pretty_print=True)



#!/bin/env python

# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import sys
import os
import textwrap

from setuptools import setup, find_packages

VERSION = '0.5.1'


# -----------------------------------------------------------------------------

try:
    import stackless  # noqa: F401
except ImportError:
    print "Warning: you are installing Nagare on CPython instead of Stackless Python (http://www.stackless.com) or PyPy (http://pypy.org)."
    print "         Without 'continuation', the 'Component.call()' method will not be available."

if sys.version_info < (2, 5, 2):
    print 'The version of Python must be 2.5.2 or more'
    sys.exit(-2)

# -----------------------------------------------------------------------------

f = open(os.path.join(os.path.dirname(__file__), 'doc', 'features.txt'))
long_description = f.read()
f.close()

setup(
    name='nagare',
    version=VERSION,
    author='Alain Poirier',
    author_email='alain.poirier@net-ng.com',
    description='Nagare Python web framework',
    long_description=textwrap.dedent("""
    Description
    ^^^^^^^^^^^

    %s

    Installation
    ============

    For a standard installation, read the `quickstart <http://www.nagare.org/doc/quickstart>`_ document.

    Read `framework installation <http://www.nagare.org/doc/framework_installation>`_ to run the
    latest development version from the `Github repository <https://github.com/nagareproject/core>`_
    or to create a Nagare developer installation.
    """) % long_description,
    license='BSD',
    keywords='web wsgi framework sqlalchemy elixir seaside continuation ajax stackless',
    url='http://www.nagare.org',
    download_url='http://www.nagare.org/download',
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['nagare/static/*']},
    namespace_packages=('nagare',),
    zip_safe=False,
    dependency_links=('http://www.nagare.org/download/',),
    install_requires=(
        'PEAK-Rules', 'ConfigObj', 'lxml', 'WebOb<1.8',
        'Paste', 'flup==1.0.3.dev-20110405', 'python-memcached'
    ),
    message_extractors={'nagare': [
        ('test/**', 'ignore', None),
        ('**.py', 'python', None),
    ]},
    extras_require={
        'debug': ('WebError',),
        'database': ('SQLAlchemy>0.5.8', 'Elixir'),
        'doc': ('sphinx', 'sphinx_rtd_theme<0.3'),
        'i18n': ('Babel>=2.5.3', 'pytz'),
        'test': (
            'SQLAlchemy>0.5.8', 'Elixir',
            'Babel>=2.5.3', 'pytz',
            'nose'
        ),
        'full': (
            'WebError',
            'SQLAlchemy>0.5.8', 'Elixir',
            'sphinx', 'sphinx_rtd_theme<0.3',
            'Babel>=2.5.3', 'pytz',
            'nose'
        ),
    },
    test_suite='nose.collector',
    entry_points='''
        [console_scripts]
        nagare-admin = nagare.admin.command:run

        [nagare.commands]
        info = nagare.admin.info:Info
        serve = nagare.admin.serve:Serve
        serve-module = nagare.admin.serve_module:Serve
        create-app = nagare.admin.create:Create
        create-db = nagare.admin.db:DBCreate
        drop-db = nagare.admin.db:DBDrop
        shell = nagare.admin.shell:Shell
        batch = nagare.admin.shell:Batch
        create-rules = nagare.admin.create_rules:CreateRules

        [nagare.publishers]
        standalone = nagare.publishers.standalone_publisher:Publisher
        threaded = nagare.publishers.standalone_publisher:Publisher
        fastcgi = nagare.publishers.fcgi_publisher:Publisher
        fapws3 = nagare.publishers.fapws_publisher:Publisher
        eventlet = nagare.publishers.eventlet_publisher:Publisher

        [nagare.sessions]
        standalone = nagare.sessions.memory_sessions:SessionsWithPickledStates
        pickle = nagare.sessions.memory_sessions:SessionsWithPickledStates
        memcache = nagare.sessions.memcached_sessions:Sessions

        [nagare.applications]
        admin = nagare.admin.admin_app:app

        [nagare.admin]
        info = nagare.admin.interface.info:Admin
        apps = nagare.admin.interface.applications:Admin
        ''',
    classifiers=(
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Environment :: Web Environment',
        'Operating System :: Microsoft :: Windows :: Windows NT/2000',
        'Operating System :: Unix',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    )
)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``shell`` and ``batch`` administrative commands

The ``shell`` command launches an interactive Python shell.
The ``batch`` command execute Python statements from a file.

In both cases:

  - the global variable ``apps`` is a dictionary of application name -> application object
  - the global variable ``session`` is the database session
  - the metadata of the applications are activated
"""

import __builtin__
import sys
import os
import code
import pkg_resources

from nagare import database, log, local
from nagare.admin import util, reference, command


def activate_applications(cfgfiles, debug, error):
    """Initialize applications

    In:
      - ``cfgfile`` -- paths to application configuration files or names of
        registered applications
      - ``debug`` -- enable the display of the generated SQL statements
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - database session
      - {application name -> application object}
    """
    # Configure the local service
    local.worker = local.Process()
    local.request = local.Process()

    configs = []

    # Merge all the ``[logging]]`` section of all the applications
    for cfgfile in cfgfiles:
        # Read the configuration file of the application
        (cfgfile, app, dist, aconf) = util.read_application(cfgfile, error)
        configs.append((cfgfile, app, dist, aconf))

        log.configure(aconf['logging'].dict(), aconf['application']['name'])

    # Configure the logging service
    log.activate()

    apps = {}

    # For each application given, activate its metadata and its logger
    for (cfgfile, app, dist, aconf) in configs:
        name = aconf['application']['name']

        log.set_logger('nagare.application.' + name)

        requirement = None if not dist else pkg_resources.Requirement.parse(dist.project_name)
        data_path = None if not requirement else pkg_resources.resource_filename(requirement, '/data')

        (apps[name], databases) = util.activate_WSGIApp(app, cfgfile, aconf, error, data_path=data_path, debug=debug)

        for (database_settings, populate) in databases:
            database.set_metadata(*database_settings)

    session = database.session
    session.begin()
    return session, apps


def activate_application(cfgfile, debug, error):
    """Initialize one application

    In:
      - ``cfgfile`` -- path to an application configuration file or name of
        a registered application
      - ``debug`` -- enable the display of the generated SQL statements
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - database session
      - application name
      - application object
    """
    session, apps = activate_applications([cfgfile], debug, error)
    return (session,) + apps.items()[0]


def create_globals(cfgfiles, debug, error):
    """Return a namespace with the initialized applications

    In:
      - ``cfgfile`` -- paths to application configuration files or names of
        registered applications
      - ``debug`` -- enable the display of the generated SQL statements
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - the namespace with the ``apps`` and ``session`` variables defined
    """
    session, apps = activate_applications(cfgfiles, debug, error)
    return {'session': session, 'apps': apps}


# -----------------------------------------------------------------------------

def set_shell_options(optparser):
    optparser.usage += ' [application ...]'

    optparser.add_option('-d', '--debug', action='store_const', const=True, default=False, dest='debug', help='debug mode for the database engine')
    optparser.add_option('--plain', action='store_const', const=False, default=True, dest='ipython', help='launch a plain Python interpreter instead of IPython')


class IPythonShellV1(object):
    """A IPython < 0.11 interpreter
    """
    def __init__(self, ipython, banner, app_names, ns):
        """Initialisation

        In:
          - ``ipython`` -- the ``IPython`` module
          - ``banner`` -- banner to display
          - ``app_names`` -- names of the activated applications
          - ``ns`` -- the namespace with the ``apps`` and ``session`` variables defined
        """
        prompt = '[%s]' % app_names[0] if len(app_names) == 1 else ''
        self.shell = ipython.Shell.IPShell(argv=['--prompt_in1=Nagare%s [\\#]: ' % prompt], user_ns=ns)
        self.banner = banner

    def __call__(self):
        """Launch the interpreter
        """
        self.shell.mainloop(1, self.banner)


class IPythonShellV2(object):
    """A IPython >= 0.11 interpreter
    """
    def __init__(self, ipython, banner, app_names, ns):
        """Initialisation

        In:
          - ``ipython`` -- the ``IPython`` module
          - ``banner`` -- banner to display
          - ``app_names`` -- names of the activated applications
          - ``ns`` -- the namespace with the ``apps`` and ``session`` variables defined
        """
        config = ipython.config.loader.Config()
        prompt = '[%s]' % app_names[0] if len(app_names) == 1 else ''
        config.PromptManager.in_template = 'Nagare%s [\\#]: ' % prompt

        self.shell = getattr(ipython, 'frontend', ipython).terminal.embed.InteractiveShellEmbed(config=config, user_ns=ns, banner1=banner)
        self.shell.confirm_exit = False

    def __call__(self):
        """Launch the interpreter
        """
        self.shell()


class IPythonShellV3(object):
    """A IPython >= 5.0 interpreter
    """
    def __init__(self, ipython, banner, app_names, ns):

        class NagarePrompts(ipython.terminal.prompts.Prompts):
            def in_prompt_tokens(self, cli=None):
                prompt = '[%s]' % app_names[0] if len(app_names) == 1 else ''
                return [
                    (ipython.terminal.prompts.Token.Prompt, 'Nagare%s [' % prompt),
                    (ipython.terminal.prompts.Token.PromptNum, str(self.shell.execution_count)),
                    (ipython.terminal.prompts.Token.Prompt, ']: '),
                ]

        self.shell = ipython.terminal.embed.InteractiveShellEmbed(banner1=banner, user_ns=ns, confirm_exit=False)
        self.shell.prompts = NagarePrompts(self.shell)

    def __call__(self):
        self.shell()


class PythonShell(code.InteractiveConsole):
    """A plain Python interpreter
    """
    def __init__(self, banner, app_names, ns):
        """Initialisation

        In:
          - ``banner`` -- banner to display
          - ``app_names`` -- names of the activated applications
          - ``ns`` -- the namespace with the ``apps`` and ``session`` variables defined
        """
        code.InteractiveConsole.__init__(self, ns)
        self.banner = banner
        self.prompt = '[%s]' % app_names[0] if len(app_names) == 1 else ''

    def raw_input(self, prompt):
        return code.InteractiveConsole.raw_input(self, 'nagare' + self.prompt + prompt)

    def __call__(self):
        self.interact(self.banner)


class PythonShellWithHistory(PythonShell):
    """A plain Python interpreter with a readline history
    """
    def __call__(self, readline):
        """Launch the interpreter

        In:
          - ``readline`` -- the ``readline`` module
          - ``banner`` -- banner to display
        """
        # Set completion on TAB and a dedicated commands history file
        readline.parse_and_bind('tab: complete')

        history_path = os.path.expanduser('~/.nagarehistory')

        if os.path.exists(history_path):
            readline.read_history_file(history_path)

        readline.set_history_length(200)

        PythonShell.__call__(self)

        readline.write_history_file(history_path)


def create_python_shell(ipython, banner, app_names, ns):
    """Shell factory

    Create a shell according to the installed modules (``readline`` and ``ipython``)

    In:
      - ``ipython`` -- does the user want a IPython shell?
      - ``banner`` -- banner to display
      - ``app_names`` -- names of the activated applications
      - ``ns`` -- the namespace with the ``apps`` and ``session`` variables defined
    """
    if ipython:
        try:
            import IPython
        except ImportError:
            pass
        else:
            ipython_version = pkg_resources.parse_version(IPython.__version__)
            if ipython_version < pkg_resources.parse_version('0.11'):
                shell_factory = IPythonShellV1
            elif ipython_version < pkg_resources.parse_version('5.0'):
                shell_factory = IPythonShellV2
            else:
                shell_factory = IPythonShellV3

            shell_factory(IPython, banner, app_names, ns)()
            return

    try:
        import readline

        PythonShellWithHistory(banner, app_names, ns)(readline)
    except ImportError:
        PythonShell(banner, app_names, ns)()


def shell(parser, options, args):
    """Launch an interactive shell

    In:
      - ``parser`` -- the ``optparse.OptParser`` object used to parse the configuration file
      - ``options`` -- options in the command line
      - ``args`` -- arguments in the command line

    The arguments are a list of names of registered applications
    or paths to applications configuration files.
    """
    ns = create_globals(args, options.debug, parser.error)
    ns['__name__'] = '__console__'

    banner = "Python %s on %s\nNagare %s\n\nVariables 'apps' and 'session' are available" % (
        sys.version,
        sys.platform,
        pkg_resources.get_distribution('nagare').version
    )

    create_python_shell(options.ipython, banner, [app.name for app in ns['apps'].values()], ns)


class Shell(command.Command):
    desc = 'Launch a shell'

    set_options = staticmethod(set_shell_options)
    run = staticmethod(shell)


# -----------------------------------------------------------------------------

def set_batch_options(optparser):
    optparser.usage += ' <application> <file.py>'

    optparser.add_option('-d', '--debug', action='store_const', const=True, default=False, dest='debug', help='debug mode for the database engine')

    if not sys.argv[3:]:
        return None

    for (i, option) in enumerate(sys.argv[3:]):
        if not option.startswith('-'):
            break

    return sys.argv[:i + 4]


def batch(parser, options, args):
    """Execute Python statements a file

    In:
      - ``parser`` -- the ``optparse.OptParser`` object used to parse the configuration file
      - ``options`` -- options in the command lines
      - ``args`` -- arguments in the command lines

    The arguments are the name of a registered applications, or the path to
    an applications configuration file, followed by the paths of a file to
    execute
    """
    if not args:
        parser.error('No application given')

    if len(args) == 1:
        parser.error('No file given')

    for (i, option) in enumerate(sys.argv[3:]):
        if not option.startswith('-'):
            break

    del sys.argv[:i + 3]

    ns = create_globals(args[:1], options.debug, parser.error)
    __builtin__.__dict__.update(ns)

    reference.load_file(args[1], None)


class Batch(command.Command):
    desc = 'Execute Python statements from a file'

    set_options = staticmethod(set_batch_options)
    run = staticmethod(batch)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Generate the content to return to the browser"""

from lxml import etree

import peak.rules
import lxml.html

from nagare.namespaces import xml, xhtml_base


@peak.rules.abstract
def serialize(output, content_type, doctype, declaration):
    """Generic method to generate the content for the browser

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    return content_type, str(output)


@peak.rules.when(serialize, (xhtml_base._HTMLTag,))
def serialize(next_method, output, content_type, doctype, declaration):
    """Generic method to generate a (X)HTML text from a tree

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    if 'xmlns' in output.attrib:
        # Let ``lxml`` generate the correct namespaces
        del output.attrib['xmlns']

    if content_type == 'application/xhtml+xml':
        # The browser accepts XHTML
        output = next_method(output, content_type, doctype, declaration)[1]
    else:
        # The browser only accepts HTML
        lxml.html.xhtml_to_html(output)

        output = output.write_htmlstring(pretty_print=True, doctype=doctype if declaration else None)

    return content_type, output


@peak.rules.when(serialize, (xml._Tag,))
def serialize(next_method, output, content_type, doctype, declaration):
    """Generic method to generate a XML text from a tree

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    if content_type == 'text/html':
        output = next_method(output, content_type, doctype, declaration)[1]
    else:
        output = output.write_xmlstring(xml_declaration=declaration, doctype=doctype if declaration else None)

    return content_type, output


@peak.rules.when(serialize, (etree._Element,))
def serialize(output, content_type, doctype, declaration):
    """Generic method to generate a XML text from a tree

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    if content_type == 'text/html':
        lxml.html.xhtml_to_html(output)
        method = 'html'
        pretty_print = True
    else:
        method = 'xml'
        pretty_print = False

    return (
        content_type,
        etree.tostring(
            output, encoding='utf-8', method=method, pretty_print=pretty_print,
            xml_declaration=declaration, doctype=doctype if declaration else None
        )
    )


@peak.rules.when(serialize, (str,))
def serialize(output, content_type, doctype, declaration):
    """Generic method to generate a text (or a binary content)

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    return content_type or 'text/plain', output


@peak.rules.when(serialize, (unicode,))
def serialize(output, content_type, doctype, declaration):
    """Generic method to generate a text from unicode

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    return serialize(output.encode('utf-8'), content_type, doctype, declaration)


@peak.rules.when(serialize, ((list, tuple),))
def serialize(output, content_type, doctype, declaration):
    """Generic method to generate a text from a list or a tuple

    In:
      - ``output`` -- the rendered content
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple (content_type, content)
    """
    if not output:
        return content_type, ''

    first = serialize(output[0], content_type, doctype, declaration)[1]
    second = ''.join(serialize(e, content_type, doctype, False)[1] for e in output[1:])

    return content_type, first + second



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Helpers to validate form datum"""

from nagare import var


class Property(var.Var):
    """An editor property

    An editor property has:

      - a current value: this value is always set
      - a valid value: this value is only set if the validating function has validated it
      - a validating function: a function that raise an exception if the value
        is invalid else, return the value to set
      - the message of the last validation error or ``None`` if the current value
        is valid
    """
    def __init__(self, v=None):
        """Initialisation

        Members are:

          - ``self.input`` -- the current value
          - ``self.value`` -- the last valid value
          - ``self.error`` -- the last error message
        """
        super(Property, self).__init__(v)

        self.value = v
        self.error = None

    def _validate(self, input):
        """Default validation function

        In:
          - ``input`` -- value to validate

        Return
          - ``input``
        """
        return input

    def validate(self, f):
        """Set the validating function

        In:
          - ``f`` -- the validating function. Called with the value to
            validate. Raise the exception ``ValueError`` is the value is
            invalid. Return the value to be set (i.e a validation function
            can do conversions too)

        Return:
          - ``self``
        """
        self._validate = f
        return self

    def set(self, input):
        """Set the values

        Always set the current value but only the valid value if the validating
        function return OK

        In:
          - ``input`` -- the input string or ``cgi.FieldStorage`` object
        """

        if not hasattr(input, 'file'):
            super(Property, self).set(input)

        try:
            self.value = self._validate(input)  # Call the validating function
            if callable(self.value):
                self.value = self.value()
            self.error = None
        except ValueError, data:
            self.error = data[0]

    def commit(self, o, name):
        """Set the attribute ``name`` of the object ``o`` with the valid value

        In:
          - ``o`` -- target object
          - ``name`` -- name of the attribute to set in the target object
        """
        setattr(o, name, self.value)


class Editor(object):
    """An editor object act as a buffer between the datum received from a
    form an the target object.

    An editor object has a set of editor properties, each with a possible
    validating function, and will modified the target object only if all
    the validating functions are passed ok.
    """
    def __init__(self, target, properties_to_create=()):
        """Initialisation

        Create the set of properties from some attributes of the target object

        In:
          - ``target`` - the target object
          - ``properties_to_create`` -- name of the attritbutes to read from the
            target object, then to write into
        """
        self.target = target

        for name in properties_to_create:
            # Create a property with the same name and value than the attribute
            # of the target object
            setattr(self, name, Property(getattr(target, name)))

    def is_validated(self, properties_to_validate):
        """Check the validity of a set of properties

        In:
          - ``properties_to_validate`` -- name of the properties to validate

        Return
          - a boolean
        """
        return all([getattr(self, name).error is None for name in properties_to_validate])

    def commit(self, properties_to_commit=(), properties_to_validate=()):
        """Write back the value of the property to the target object, only if
        they are all valid

        In:
          - ``properties_to_commit`` -- names of properties to check then to
            write to the target object
          - ``properties_to_validate`` -- additional names of properties to check

        Return:
          - are all the properties valid ?
        """
        validated = self.is_validated(set(properties_to_commit) | set(properties_to_validate))

        if validated:
            for name in properties_to_commit:
                getattr(self, name).commit(self.target, name)

        return validated



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from nagare import local, wsgi
from nagare.sessions import ExpirationError, common

local.request = local.Process()


def create_environ():
    return {
        'REQUEST_METHOD': 'GET',
        'SCRIPT_NAME': '/app',
        'PATH_INFO': '/',
        'QUERY_STRING': '_s=10&_c=42',
        'SERVER_PROTOCOL': 'HTTP/1.0',
        'SERVER_NAME': 'localhost',
        'SERVER_PORT': 8080,
        'wsgi.url_scheme': 'http'
    }


class Response(dict):
    def __init__(self):
        self.status_code = None
        self.status = None

    def __call__(self, status, headers):
        (self.status_code, self.status) = status.split(' ', 1)
        self.status_code = int(self.status_code)

        self.update(dict(headers))
        assert len(self) == len(headers)

    def __repr__(self):
        return '%d %s - %s' % (self.status_code, self.status, super(Response, self).__repr__())


class SessionManager(common.Sessions):
    pass


class Lock(object):
    acquire = release = lambda self: None


class ExpiredSessionManager(common.Sessions):
    def get_lock(self, session_id):
        return Lock()

    def get_state(self, request, response, use_same_state):
        assert self._get_ids(request) == (10, 42)
        raise ExpirationError()


class App(wsgi.WSGIApp):
    def __init__(self, session_manager=SessionManager(local.DummyLock)):
        super(App, self).__init__(lambda: None)
        self.sessions = session_manager


def process_request(app=None, environ={}, **kw):
    if app is None:
        app = App()

    env = create_environ()
    env.update(environ)
    env.update(kw)

    r = Response()
    app(env, r)
    return r


def test_request_validity2():
    """Request - invalid url"""
    r = process_request(PATH_INFO='')
    assert (r.status_code == 301) and r['Location'] == 'http://localhost:8080/app/?_s=10&_c=42'


def test_bad_session():
    """Request - session expired"""
    r = process_request(App(session_manager=ExpiredSessionManager(local.DummyLock)))
    assert (r.status_code == 301) and r['Location'] == 'http://localhost:8080/app/'



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Base class of all the renderers"""

import random


class Renderer(object):
    def new(self):
        """Create a new renderer from the same type of this renderer
        """
        return self.__class__(self)

    def start_rendering(self, comp, model):
        """Method calls before to render a component

        In:
          - ``comp`` -- the component to render
          - ``model`` -- the view to render
        """
        pass

    def end_rendering(self, output):
        """Method call after the rendering of the component

        In:
          - ``output`` -- the rendering tree
        """
        return output

    def generate_id(self, prefix=''):
        """Generate a random id

        In:
          - ``prefix`` -- prefix of the generated id
        """
        return prefix + str(random.randint(10000000, 99999999))

    def __reduce__(self):
        """Prevent a renderer to be pickled

        Prevent the common error of capturing the renderer into the closure
        of a callback
        """
        raise TypeError("can't pickle Renderer objects (are you using a renderer object in a callback ?)")



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from nagare import editor, validator


class MyEditor(editor.Editor):
    fields = ('name', 'age')

    def __init__(self, source):
        super(MyEditor, self).__init__(source, self.fields)

        self.name.validate(lambda v: validator.to_string(v, strip=True).not_empty().to_string())
        self.age.validate(lambda v: validator.to_int(v).lesser_than(50).greater_than(0).to_int())

    def commit(self):
        super(MyEditor, self).commit(self.fields)

    def set_values(self, name, age):
        self.name(name)
        self.age(age)


class MyApp:
    def __init__(self):
        self.name = 'Foo'
        self.age = 5


def test1():
    """ Form - validation OK """
    o = MyApp()

    assert o.name == 'Foo'
    assert o.age == 5

    editor = MyEditor(o)

    assert o.name == 'Foo'
    assert o.age == 5

    editor.set_values('Bar', '4')

    # Editor values are changed
    assert editor.name.input == editor.name() == 'Bar'
    assert editor.age.input == editor.age() == '4'
    assert editor.age.value == 4

    # App values aren't changed
    assert o.name == 'Foo'
    assert o.age == 5

    editor.commit()

    # App values are changed
    assert o.name == 'Bar'
    assert o.age == 4


def test2():
    """ Form - validation KO """
    o = MyApp()

    assert o.name == 'Foo'
    assert o.age == 5

    editor = MyEditor(o)

    assert o.name == 'Foo'
    assert o.age == 5

    editor.set_values('Bar', '1000')

    # Editor values are changed
    assert editor.name.input == editor.name() == 'Bar'
    assert editor.age.input == editor.age() == '1000'
    assert editor.age.value == 5

    # App values aren't changed
    assert o.name == 'Foo'
    assert o.age == 5

    editor.commit()

    # App values aren't changed
    assert o.name == 'Foo'
    assert o.age == 5


class MyStringEditor(editor.Editor):
    fields = ('name',)

    def __init__(self, source):
        super(MyStringEditor, self).__init__(source, self.fields)

    def commit(self):
        super(MyStringEditor, self).commit(self.fields)

    def set_values(self, name):
        self.name(name)


class MyStringEditor1(MyStringEditor):
    def __init__(self, source):
        super(MyStringEditor1, self).__init__(source)

        self.name.validate(lambda v: validator.to_string(v, strip=True).not_empty())


def test4():
    """ Form - test string validators - not_empty """
    editor = MyStringEditor1(MyApp())

    # Validation OK
    editor.set_values('abcd')
    assert editor.name.error is None
    assert editor.name.input == editor.name() == 'abcd'
    assert editor.name.value == 'abcd'

    # Validation KO
    editor.set_values('')
    assert editor.name.error == "Can't be empty"
    assert editor.name.input == editor.name() == ''
    assert editor.name.value == 'abcd'


class MyStringEditor2(MyStringEditor):
    def __init__(self, source):
        super(MyStringEditor2, self).__init__(source)
        self.name.validate(lambda v: validator.to_string(v, strip=True).match(r'^[a-d]+$', msg="test5 - Incorrect format"))


def test5():
    """ Form - test string validators - match """
    editor = MyStringEditor2(MyApp())

    # Validation OK
    editor.set_values('abab')
    assert editor.name.error is None
    assert editor.name.input == editor.name() == 'abab'
    assert editor.name.value == 'abab'

    # Validation KO
    editor.set_values('abrab')
    assert editor.name.error == "test5 - Incorrect format"
    assert editor.name.input == editor.name() == 'abrab'
    assert editor.name.value == 'abab'


class MyStringEditor3(MyStringEditor):
    def __init__(self, source):
        super(MyStringEditor3, self).__init__(source)
        self.name.validate(lambda v: validator.to_string(v, strip=True).shorter_than(5, msg="test6 - Length must be shorter than %(max)d characters"))


def test6():
    """ Form - test string validators - shorter_than """
    editor = MyStringEditor3(MyApp())

    # Validation OK
    editor.set_values('1234')
    assert editor.name.error is None
    assert editor.name.input == editor.name() == '1234'
    assert editor.name.value == '1234'

    # Validation KO
    editor.set_values('123456')
    assert editor.name.error == "test6 - Length must be shorter than 5 characters"
    assert editor.name.input == editor.name() == '123456'
    assert editor.name.value == '1234'


class MyStringEditor4(MyStringEditor):
    def __init__(self, source):
        super(MyStringEditor4, self).__init__(source)
        self.name.validate(lambda v: validator.to_string(v, strip=True).length_equal(5, msg="test7 - Length must be %(len)d characters"))


def test7():
    """ Form - test string validators - length_equal  """
    editor = MyStringEditor4(MyApp())

    # Validation OK
    editor.set_values('12345')
    assert editor.name.error is None
    assert editor.name.input == editor.name() == '12345'
    assert editor.name.value == '12345'

    # Validation KO
    editor.set_values('123456')
    assert editor.name.error == "test7 - Length must be 5 characters"
    assert editor.name.input == editor.name() == '123456'
    assert editor.name.value == '12345'

    # Validation KO
    editor.set_values('1234')
    assert editor.name.error == "test7 - Length must be 5 characters"
    assert editor.name.input == editor.name() == '1234'
    assert editor.name.value == '12345'


class MyStringEditor5(MyStringEditor):
    def __init__(self, source):
        super(MyStringEditor5, self).__init__(source)
        self.name.validate(lambda v: validator.to_string(v, strip=True).longer_than(5).shorter_or_equal_than(8).not_empty().match(r'^[1-9]+$'))


def test8():
    """ Form - test string validators - multiple validators """
    editor = MyStringEditor5(MyApp())

    # Validation OK
    editor.set_values('123456')
    assert editor.name.error is None
    assert editor.name.input == editor.name() == '123456'
    assert editor.name.value == '123456'

    # Validation KO
    editor.set_values('')
    assert editor.name.error is not None
    assert editor.name.input == editor.name() == ''
    assert editor.name.value == '123456'

    # Validation KO
    editor.set_values('1234')
    assert editor.name.error is not None
    assert editor.name.input == editor.name() == '1234'
    assert editor.name.value == '123456'

    # Validation KO
    editor.set_values('123456789')
    assert editor.name.error is not None
    assert editor.name.input == editor.name() == '123456789'
    assert editor.name.value == '123456'

    # Validation KO
    editor.set_values('abcdefg')
    assert editor.name.error is not None
    assert editor.name.input == editor.name() == 'abcdefg'
    assert editor.name.value == '123456'


def test9():
    p = editor.Property()
    assert (p.input is None) and (p() is None) and (p.value is None) and (p.error is None)


def test10():
    p = editor.Property(5)
    assert (p.input == p() == p.value == 5) and (p.error is None)


def check(v):
    if v > 10:
        raise ValueError('invalid')
    return v


def test11():
    p = editor.Property(5).validate(check)
    assert (p.input == p() == p.value == 5) and (p.error is None)


def test12():
    p = editor.Property(15).validate(check)
    assert (p.input == p() == p.value == 15) and (p.error is None)


def test13():
    p = editor.Property().validate(check)
    p(5)
    assert (p.input == p() == p.value == 5) and (p.error is None)


def test14():
    p = editor.Property().validate(check)
    p(15)
    assert (p.input == p() == 15) and (p.value is None) and (p.error == 'invalid')



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Set of validating objects

Suitable to be the validating functions of ``editor.property`` objects
"""

import re
import functools

from nagare import i18n, partial


_L = partial.Partial(i18n._L, domain='nagare')


class DualCallable(type):
    """"A hackish metaclass to allow both direct and deferred calls of methods

    For compatibility with the old and new way to built a validation chain.

    Examples:

      - Old validation with direct calls: valid = lambda v: IntValidator(v).greater_than(10)
      - New validation with lazy calls: valid = IntValidator().greater_than(10)
    """
    def __new__(cls, name, bases, ns):
        """Class Initialization

        Wrap all the methods in ``ns``

        In:
          - ``name`` -- name of the class to create
          - ``bases`` -- base classes of the class to create
          - ``ns`` -- namespace of the class to create

        Return:
          - a new class
        """
        for method_name, method in ns.iteritems():
            if callable(method) and (method_name == '__init__' or not method_name.startswith('_')):

                def _(m):
                    def f(self, *args, **kw):
                        return self._call_or_defer(m, *args, **kw)
                    functools.update_wrapper(f, m)
                    return f

                ns[method_name] = _(method)

        return super(DualCallable, cls).__new__(cls, name, bases, ns)


class ValidatorBase(object):
    """"A hackish base class to allow both direct and deferred calls of methods

    For compatibility with the old and new way to built a validation chain.

    Examples:

      - Old validation with direct calls: valid = lambda v: IntValidator(v).greater_than(10)
      - New validation with lazy calls: valid = IntValidator().greater_than(10)
    """

    __metaclass__ = DualCallable

    def __new__(cls, v=None, *args, **kw):
        """Method called before ``__init__``

        If a ``v`` value is passed, all the methods will be directly called
        else, the method calls will be recorded and called later

        In:
          - ``v`` -- optional value to validate
          - ``args``, ``kw`` -- ``__init__`` parameters

        Return:
          - an instance
        """
        o = super(ValidatorBase, cls).__new__(cls)
        o._defer = v is None
        o._methods_chain = []
        return o

    def _call_or_defer(self, _method, *args, **kw):
        """Directly call or record the call to a method

        In:
          - ``_method`` -- method to call (keyword parameter)
          - ``args``, ``kw`` -- parameters of ``_method``
        """
        if self._defer:
            # Record the call to the method
            self._methods_chain.append((_method, args, kw))
            return None if _method.__name__ == '__init__' else self

        # Directly call the method
        return _method(self, *args, **kw)

    def __call__(self, v=None):
        """Return the already validated value or call now all the deferred methods

        In:
          - ``v`` -- optional value to validate

        Return:
          - the final result of all the calls
        """
        if self._defer:
            self._defer = False

            try:
                method, args, kw = self._methods_chain[0]
                method(self, v, *args, **kw)

                for method, args, kw in self._methods_chain[1:]:
                    method(self, *args, **kw)
            finally:
                self._defer = True

        return self.value


class Validator(ValidatorBase):
    """Base class for the validation objects
    """
    def __init__(self, v, strip=False, rstrip=False, lstrip=False, chars=None):
        """Initialization

        This object only do conversions, possibly removing characters at the
        beginning / end of the value

        In:
          - ``v`` -- value to validate
          - ``strip`` -- remove the characters at the beginning and the end
          - ``rstrip`` -- remove the characters at the end
          - ``lstrip`` -- remove the characters at the beginning
          - ``chars`` -- list of characters to removed, spaces by default
        """
        if not isinstance(v, basestring):
            raise ValueError('Input must be a string')

        if strip:
            v = v.strip(chars)

        if rstrip:
            v = v.rstrip(chars)

        if lstrip:
            v = v.lstrip(chars)

        self.value = v


class IntValidator(Validator):
    """Conversion and validation of integers
    """
    def __init__(self, v, base=10, msg=i18n._L('Must be an integer'), *args, **kw):
        """Initialisation

        Check that the value is an integer

        In:
          - ``v`` -- value to validate
        """
        super(IntValidator, self).__init__(v, *args, **kw)

        try:
            self.value = int(self.value, base)
        except (ValueError, TypeError):
            raise ValueError(msg)

    to_int = Validator.__call__

    def lesser_than(self, max, msg=_L('Must be lesser than %(max)d')):
        """Check that the value is lesser than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if self.value < max:
            return self

        raise ValueError(msg % {'value': self.value, 'max': max})

    def lesser_or_equal_than(self, max, msg=_L('Must be lesser or equal than %(max)d')):
        """Check that the value is lesser or equal than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if self.value <= max:
            return self

        raise ValueError(msg % {'value': self.value, 'max': max})

    def greater_than(self, min, msg=_L('Must be greater than %(min)d')):
        """Check that the value is greater than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if self.value > min:
            return self

        raise ValueError(msg % {'value': self.value, 'min': min})

    def greater_or_equal_than(self, min, msg=_L('Must be greater or equal than %(min)d')):
        """Check that the value is greater or equal than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if self.value >= min:
            return self

        raise ValueError(msg % {'value': self.value, 'min': min})


class StringValidator(Validator):
    """Conversion and validation of string
    """
    to_string = Validator.__call__

    def to_int(self, base=10):
        """Return the value, converted to an integer

        In:
          - ``base`` -- base for the conversion

        Return:
          - the integer value
        """
        self.value = int(self.value, base=base)
        return self.value

    def not_empty(self, msg=_L("Can't be empty")):
        """Check that the value is not empty

        In:
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) != 0:
            return self

        raise ValueError(msg)

    def match(self, r, msg=_L('Incorrect format')):
        """Check that the value respects a format given as a regexp

        In:
          - ``r`` -- the regexp
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if re.match(r, self.value):
            return self

        raise ValueError(msg % {'value': self.value})

    def shorter_than(self, max, msg=_L('Length must be shorter than %(max)d characters')):
        """Check that the value is shorter than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) < max:
            return self

        raise ValueError(msg % {'value': self.value, 'max': max})

    def shorter_or_equal_than(self, max, msg=_L('Length must be shorter or equal than %(max)d characters')):
        """Check that the value is shorter or equal than a limit

        In:
          - ``max`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) <= max:
            return self

        raise ValueError(msg % {'value': self.value, 'max': max})

    def length_equal(self, v, msg=_L('Length must be %(len)d characters')):
        """Check that the value has an exact length

        In:
          - ``v`` -- the length
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) == v:
            return self

        raise ValueError(msg % {'value': self.value, 'len': v})

    def longer_than(self, min, msg=_L('Length must be longer than %(min)d characters')):
        """Check that the value is longer than a limit

        In:
          - ``min`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) > min:
            return self

        raise ValueError(msg % {'value': self.value, 'min': min})

    def longer_or_equal_than(self, min, msg=_L('Length must be longer or equal than %(min)d characters')):
        """Check that the value is longer or equal than a limit

        In:
          - ``min`` -- the limit
          - ``msg`` -- message to raise

        Return:
          - ``self``
        """
        if len(self.value) >= min:
            return self

        raise ValueError(msg % {'value': self.value, 'min': min})

    def isalnum(self, msg=_L('Some characters are not alphanumeric')):
        if self.value.isalnum():
            return self

        raise ValueError(msg % {'value': self.value})

    def isalpha(self, msg=_L('Some characters are not alphabetic')):
        if self.value.isalpha():
            return self

        raise ValueError(msg % {'value': self.value})

    def isdigit(self, msg=_L('Some characters are not digits')):
        if self.value.isdigit():
            return self

        raise ValueError(msg % {'value': self.value})

    def islower(self, msg=_L('Some characters are not lowercase')):
        if self.value.islower():
            return self

        raise ValueError(msg % {'value': self.value})

    def isupper(self, msg=_L('Some characters are not uppercase')):
        if self.value.isupper():
            return self

        raise ValueError(msg % {'value': self.value})

    def isspace(self, msg=_L('Some characters are not whitespace')):
        if self.value.isspace():
            return self

        raise ValueError(msg % {'value': self.value})


# Aliases
to_int = IntValidator
to_string = StringValidator



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``info`` administrative command

Display informations about the framework environment
"""

import sys
import pkg_resources

from nagare.admin import command


class Info(command.Command):
    """Display informations about the framework environment"""

    desc = 'Display various informations'

    @staticmethod
    def run(parser, options, args):
        """Display the informations

        In:
          - ``parser`` -- the optparse.OptParser object used to parse the configuration file
          - ``options`` -- options in the command lines
          - ``args`` -- arguments in the command lines
        """
        # For the moment, just display the Python version
        print sys.subversion[0], sys.version
        print
        print 'Nagare version', pkg_resources.get_distribution('nagare').version



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Serializable partial functions"""

import sys
import types
import pickle
import copy_reg

try:
    import stackless  # noqa: F401
except ImportError:
    def pickle_method(m):
        """Serialize a method

        In:
          - ``m`` -- method to pickle

        Return:
          - tuple to pickle (class of the method, name of the method, self)
        """
        return unpickle_method, (m.im_class, m.__name__, m.im_self)

    def unpickle_method(cls, name, o):
        """Deserialize a method

        In:
          - ``cls`` -- class of the method
          - ``name`` -- name of the method
          - ``o`` -- self

        Return:
          - the method, bound to ``o``
        """
        return getattr(o if isinstance(o, type) else cls, name).__get__(o, cls)

    copy_reg.pickle(types.MethodType, pickle_method)

    def pickle_function(f):
        """Serialize a function

        Only called when a lambda must be serialized

        In:
          - ``f`` -- function to pickle
        """
        msg = "Can't pickle %r, file \"%s\", line %d" % (f, f.func_code.co_filename, f.func_code.co_firstlineno)
        raise pickle.PicklingError(msg)

    copy_reg.pickle(types.FunctionType, pickle_function)


# -----------------------------------------------------------------------------

def max_number_of_args(nb):
    """Limit the number of positional parameters

    In:
      - ``nb`` - max number of positional parameters passed to ``f``. Other
                 positional parameters will be passed as the ``args`` tuple
    """
    def _(f):
        return lambda *args, **kw: f(*args[:nb], args=args[nb:], **kw)
    return _


# -----------------------------------------------------------------------------

class _Partial(object):
    def __init__(self, _f, *args, **kw):
        """Callable with predefined parameters

        A partial object when called will behave like ``_f`` called with the
        positional arguments ``args`` and keyword arguments ``kw``

        Like the standard ``functools.partial()`` but Serializable.

        In:
          - ``_f`` -- function to wrap
          - ``args``, ``kw`` -- ``_f`` parameters
        """
        self.f = _f
        self.args = args
        self.kw = kw

    def __call__(self, *args, **kw):
        """Call the wrapper function

        In:
          - ``args`` -- parameters added _after_ ``self.args``
          - ``kw`` -- parameters that can override ``self.kw``

        Return:
          - return of the wrapper function
        """
        k = self.kw.copy()
        k.update(kw)
        return self.f(*(self.args + args), **k)


def Partial(__f, *args, **kw):
    """Don't double wrap a ``_Partial()`` object if not needed"""
    return _Partial(__f, *args, **kw) if (not isinstance(__f, _Partial) or args or kw) else __f


# -----------------------------------------------------------------------------

class Decorator(object):
    """Use a ``_Partial()`` to decorate a function
    """
    def __init__(self, f, new_f, *args, **kw):
        """Decorate a function or a method

        In:
          - ``f`` -- function or method to decorate
          - ``new_f`` -- function that will decorate ``f``
          - ``args``, ``kw`` -- ``new_f`` parameters
        """

        # Hack:
        #  - change the name of ``f``
        #  - attach ``f`` to the global scope of its module
        #    (needed to be able to serialize it)
        m = sys.modules[f.__module__]
        f.__name__ = '__' + f.__name__
        setattr(m, f.__name__, f)

        self.f = f
        self.new_f = new_f
        self.args = args
        self.kw = kw

    def create_partial(self, o=None):
        """Create a ``_Partial()`` to call ``new_f``

        ``new_f`` will be called with the parameters:

          - ``self`` (``None`` if ``f`` is a function)
          - ``f``
          - ``args``, ``kw``
        """
        return Partial(self.new_f, o, self.f, *self.args, **self.kw)

    def __get__(self, o, cls):
        """Called when this decorator decorates a method
        """
        return self.create_partial(o)

    def __call__(self, *args, **kw):
        """Called when this decorator decorates a function
        """
        return self.create_partial()(*args, **kw)
